/*
 *  libMirage: private utility functions and helpers
 *  Copyright (C) 2025-2026 Rok Mandeljc
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#pragma once

#include <glib.h>
#include <glib-object.h> /* GCallback */


G_BEGIN_DECLS


/* Miscellaneous */
G_GNUC_INTERNAL
guint mirage_signal_handlers_disconnect_by_func (gpointer instance, GCallback func, gpointer user_data);


G_END_DECLS
